@@ -1,11 +1,12 @@
 /* gtkfilesystemmemory.c
  *
- * Copyright (C) 2005 Nokia Corporation.
+ * Copyright (C) 2005, 2006 Nokia Corporation.
+ *
+ * Contact: Marius Vollmer <marius.vollmer@nokia.com>
  *
  * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * modify it under the terms of the GNU Lesser General Public License
+ * version 2.1 as published by the Free Software Foundation.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -14,9 +15,11 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
- *
- *
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
+ * 02110-1301 USA
+ */
+
+/*
  * TODO - Before the converting functions were public there was only
  *        conversions from GtkTreePath to GtkFilePath.
  *        Most probably there is cases where to use straight conversion to
@@ -27,8 +30,6 @@
 #include <gtk/gtktreestore.h>
 
 #define GTK_FILE_SYSTEM_ENABLE_UNSUPPORTED
-#undef __GNUC__ /* This is needed because compile option -pedantic
-                   disables GNU extensions and code don't detect this */
 
 #include <gtk/gtkfilesystem.h>
 
@@ -69,9 +70,9 @@ struct _GtkFileSystemMemoryClass
 /*#define SOME_TESTING_STUFF 1 */
 
 #if PRINT_FUNCTION_NAME
-  #define DTEXT(str) fprintf( stderr, str )
+#define DTEXT(str) fprintf( stderr, "TRACE: %s\n", str )
 #else
-  #define DTEXT(str)
+#define DTEXT(str)
 #endif
 
 #define GTK_TYPE_FILE_FOLDER_MEMORY (gtk_file_folder_memory_get_type())
@@ -121,14 +122,17 @@ static GSList
 static GtkFileSystemVolume *
 gtk_file_system_memory_get_volume_for_path( GtkFileSystem *file_system,
 									                          const GtkFilePath *path );
-static GtkFileFolder *
-gtk_file_system_memory_get_folder( GtkFileSystem *file_system,
-							                     const GtkFilePath *path, GtkFileInfoType
-                                   types, GError **error );
-static gboolean
-gtk_file_system_memory_create_folder( GtkFileSystem *file_system,
-							                        const GtkFilePath *path,
-                                      GError **error );
+static GtkFileSystemHandle *
+gtk_file_system_memory_get_folder (GtkFileSystem *file_system,
+				   const GtkFilePath *path,
+				   GtkFileInfoType types,
+				   GtkFileSystemGetFolderCallback callback,
+				   gpointer data);
+static GtkFileSystemHandle *
+gtk_file_system_memory_create_folder (GtkFileSystem *file_system,
+				      const GtkFilePath *path,
+				      GtkFileSystemCreateFolderCallback callback,
+                                      gpointer data);
 static void
 gtk_file_system_memory_volume_free( GtkFileSystem  *file_system,
 						                        GtkFileSystemVolume *volume );
@@ -138,18 +142,22 @@ gtk_file_system_memory_volume_get_base_path( GtkFileSystem *file_system,
 static gboolean
 gtk_file_system_memory_volume_get_is_mounted( GtkFileSystem *file_system,
 								                              GtkFileSystemVolume *volume );
-static gboolean
-gtk_file_system_memory_volume_mount( GtkFileSystem *file_system,
-								                     GtkFileSystemVolume *volume,
-                                     GError **error );
+static GtkFileSystemHandle *
+gtk_file_system_memory_volume_mount (GtkFileSystem *file_system,
+				     GtkFileSystemVolume *volume,
+				     GtkFileSystemVolumeMountCallback callback,
+				     gpointer data);
 static gchar *
 gtk_file_system_memory_volume_get_display_name( GtkFileSystem *file_system,
 								                                GtkFileSystemVolume *volume );
+#if 0
 static GdkPixbuf *
 gtk_file_system_memory_volume_render_icon( GtkFileSystem *file_system,
 								                           GtkFileSystemVolume *volume,
                                            GtkWidget *widget, gint pixel_size, 
                                            GError **error );
+#endif
+
 static gboolean
 gtk_file_system_memory_get_parent( GtkFileSystem *file_system,
 							                     const GtkFilePath *path,
@@ -175,10 +183,12 @@ gtk_file_system_memory_uri_to_path( GtkFileSystem *file_system,
 static GtkFilePath *
 gtk_file_system_memory_filename_to_path( GtkFileSystem *file_system,
 								                         const gchar *filename );
+#if 0
 static GdkPixbuf *
 gtk_file_system_memory_render_icon( GtkFileSystem *file_system,
 							                      const GtkFilePath *path, GtkWidget *widget,
                                     gint pixel_size, GError **error );
+#endif
 static gboolean
 gtk_file_system_memory_insert_bookmark( GtkFileSystem *file_system,
 							                          const GtkFilePath *path, gint position,
@@ -205,7 +215,14 @@ static void
 gtk_file_folder_memory_finalize( GObject *object );
 
 
-
+#if 0
+static GtkFileSystemHandle *
+gtk_file_system_memory_get_info (GtkFileSystem                *file_system,
+				 const GtkFilePath            *path,
+				 GtkFileInfoType               types,
+				 GtkFileSystemGetInfoCallback  callback,
+				 gpointer                      data);
+#endif
 
 static GtkFileInfo *
 gtk_file_folder_memory_get_info( GtkFileFolder *folder, const GtkFilePath *path,
@@ -228,16 +245,197 @@ void signal_row_changed( GtkTreeModel *treemodel, GtkTreePath *path,
 void signal_row_deleted( GtkTreeModel *treemodel, GtkTreePath *path,
                          gpointer data);
 
-                         
-void
-signal_row_reference_inserted( GtkTreeModel *treemodel, GtkTreePath *path,
-                               GtkTreeIter *iter, GtkFileFolderMemory *ffm );
-void
-signal_row_reference_changed( GtkTreeModel *treemodel, GtkTreePath *path,
-                               GtkTreeIter *iter, GtkFileFolderMemory *ffm );
-void
-signal_row_reference_deleted( GtkTreeModel *treemodel, GtkTreePath *path,
-                               GtkFileFolderMemory *ffm );
+
+/* Maintaining callbacks.
+
+   All our operations are synchronous, but the clients of the
+   asynchronous GtkFileSystem API expect their callbacks to be invoked
+   from the main loop.  Thus, we queue them up for a idle handler.
+
+   We don't ever cancel a operation.
+
+   Because we don't cancel operations, we don't strictly need handles
+   except that clients might expect to get a unique handle for each
+   operation.  So we give them one.
+*/
+
+static void
+gtk_file_system_memory_cancel_operation (GtkFileSystemHandle *handle)
+{
+}
+
+enum callback_type {
+  CALLBACK_GET_FOLDER,
+  CALLBACK_CREATE_FOLDER,
+  CALLBACK_VOLUME_MOUNT
+};
+
+typedef struct {
+  GtkFileSystemHandle *handle;
+  GError *error;
+  gpointer callback_data;
+
+  enum callback_type type;
+  union {
+    struct {
+      GtkFileSystemGetFolderCallback callback;
+      GtkFileFolder *folder;
+    } get_folder;
+    struct {
+      GtkFileSystemCreateFolderCallback callback;
+      const GtkFilePath *path;
+    } create_folder;
+    struct {
+      GtkFileSystemVolumeMountCallback callback;
+      GtkFileSystemVolume *volume;
+    } volume_mount;
+  } info;
+} callback_info;
+
+static gint idle_handler_id = 0;
+static GSList *pending_callbacks = NULL;
+
+static gboolean
+execute_callbacks (gpointer unused)
+{
+  GSList *list, *l;
+
+  GDK_THREADS_ENTER ();
+
+  list = pending_callbacks;
+  pending_callbacks = NULL;
+  idle_handler_id = 0;
+
+  GDK_THREADS_LEAVE ();
+
+  for (l = list; l; l = l->next)
+    {
+      callback_info *info = (callback_info *)l->data;
+      GtkFileSystemHandle *handle = info->handle;
+      
+      switch (info->type)
+	{
+	case CALLBACK_GET_FOLDER:
+	  info->info.get_folder.callback (handle,
+					  info->info.get_folder.folder, 
+					  info->error,
+					  info->callback_data);
+	  break;
+	case CALLBACK_CREATE_FOLDER:
+	  info->info.create_folder.callback (handle,
+					     info->info.create_folder.path, 
+					     info->error,
+					     info->callback_data);
+	  break;
+	case CALLBACK_VOLUME_MOUNT:
+	  info->info.volume_mount.callback (handle,
+					    info->info.volume_mount.volume, 
+					    info->error,
+					    info->callback_data);
+	  break;
+	default:
+	  g_assert_not_reached ();
+	  break;
+	}
+
+      if (info->error)
+	g_error_free (info->error);
+
+      g_object_unref (handle->file_system);
+      handle->file_system = NULL;
+      g_object_unref (handle);
+
+      g_free (info);
+    }
+  
+  g_slist_free (list);
+
+  return FALSE;
+}
+
+static callback_info *
+queue_new_info (enum callback_type type,
+		GtkFileSystem *file_system,
+		GError *error,
+		gpointer callback_data)
+{
+  GtkFileSystemHandle *handle;
+  callback_info *info;
+
+  handle = g_object_new (GTK_TYPE_FILE_SYSTEM_HANDLE, NULL);
+  handle->file_system = file_system;
+  g_object_ref (file_system);
+
+  info = g_new (callback_info, 1);
+  info->type = type;
+  info->handle = handle;
+  info->error = error;
+  info->callback_data = callback_data;
+  
+  g_object_ref (handle);
+
+  GDK_THREADS_ENTER ();
+
+  pending_callbacks = g_slist_prepend (pending_callbacks, info);
+  if (idle_handler_id == 0)
+    idle_handler_id = g_idle_add (execute_callbacks, NULL);
+
+  GDK_THREADS_LEAVE ();
+
+  return info;
+}
+
+static GtkFileSystemHandle *
+queue_get_folder_callback (GtkFileSystem *file_system,
+			   GtkFileSystemGetFolderCallback callback,
+			   GtkFileFolder *folder,
+			   GError *error,
+			   gpointer callback_data)
+{
+  callback_info *info = queue_new_info (CALLBACK_GET_FOLDER,
+					file_system,
+					error,
+					callback_data);
+
+  info->info.get_folder.callback = callback;
+  info->info.get_folder.folder = folder;
+  return info->handle;
+}
+
+static GtkFileSystemHandle *
+queue_create_folder_callback (GtkFileSystem *file_system,
+			      GtkFileSystemCreateFolderCallback callback,
+			      const GtkFilePath *path,
+			      GError *error,
+			      gpointer callback_data)
+{
+  callback_info *info = queue_new_info (CALLBACK_CREATE_FOLDER,
+					file_system,
+					error,
+					callback_data);
+
+  info->info.create_folder.callback = callback;
+  info->info.create_folder.path = path;
+  return info->handle;
+}
+
+static GtkFileSystemHandle *
+queue_volume_mount_callback (GtkFileSystem *file_system,
+			     GtkFileSystemVolumeMountCallback callback,
+			     GtkFileSystemVolume *volume,
+			     GError *error,
+			     gpointer callback_data)
+{
+  callback_info *info = queue_new_info (CALLBACK_VOLUME_MOUNT,
+					file_system,
+					error,
+					callback_data);
+
+  info->info.volume_mount.callback = callback;
+  info->info.volume_mount.volume = volume;
+  return info->handle;
+}
+
 
 void
 fs_module_init( GTypeModule *module );
@@ -332,7 +530,7 @@ unescape_character (const char *scanner)
   int first_digit;
   int second_digit;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   first_digit = g_ascii_xdigit_value (scanner[0]);
   if (first_digit < 0) 
@@ -353,7 +551,7 @@ g_unescape_uri_string (const char *escaped)
   gchar *out, *result;
   int c, len;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__  );
 
   if (escaped == NULL)
     return NULL;
@@ -432,9 +630,9 @@ gtk_file_system_memory_new( void )
 {
   GtkFileSystemMemory *mfs;
 
-  mfs = g_object_new( GTK_TYPE_FILE_SYSTEM_MEMORY, NULL );
+  DTEXT( __FUNCTION__ );
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  mfs = g_object_new( GTK_TYPE_FILE_SYSTEM_MEMORY, NULL );
 
   return GTK_FILE_SYSTEM( mfs );
 }
@@ -446,7 +644,7 @@ gtk_file_system_memory_class_init( GtkFileSystemMemoryClass *class )
   GObjectClass *gobject_class = G_OBJECT_CLASS( class );
   system_parent_class = g_type_class_peek_parent( class );
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   gobject_class->finalize = gtk_file_system_memory_finalize;
 }
@@ -458,14 +656,19 @@ gtk_file_system_memory_iface_init( GtkFileSystemIface *iface )
   iface->list_volumes = gtk_file_system_memory_list_volumes;
   iface->get_volume_for_path = gtk_file_system_memory_get_volume_for_path;
   iface->get_folder = gtk_file_system_memory_get_folder;
+#if 0
+  iface->get_info = gtk_file_system_memory_get_info;
+#endif
   iface->create_folder = gtk_file_system_memory_create_folder;
+  iface->cancel_operation = gtk_file_system_memory_cancel_operation;
   iface->volume_free = gtk_file_system_memory_volume_free;
   iface->volume_get_base_path = gtk_file_system_memory_volume_get_base_path;
   iface->volume_get_is_mounted = gtk_file_system_memory_volume_get_is_mounted;
   iface->volume_mount = gtk_file_system_memory_volume_mount;
-  iface->volume_get_display_name = 
-                        gtk_file_system_memory_volume_get_display_name;
-  iface->volume_render_icon = gtk_file_system_memory_volume_render_icon;
+  iface->volume_get_display_name = gtk_file_system_memory_volume_get_display_name;
+#if 0
+  iface->volume_get_icon_name = gtk_file_system_memory_volume_get_icon_name;
+#endif
   iface->get_parent = gtk_file_system_memory_get_parent;
   iface->make_path = gtk_file_system_memory_make_path;
   iface->parse = gtk_file_system_memory_parse;
@@ -473,10 +676,13 @@ gtk_file_system_memory_iface_init( GtkFileSystemIface *iface )
   iface->path_to_filename = gtk_file_system_memory_path_to_filename;
   iface->uri_to_path = gtk_file_system_memory_uri_to_path;
   iface->filename_to_path = gtk_file_system_memory_filename_to_path;
-  iface->render_icon = gtk_file_system_memory_render_icon;
   iface->insert_bookmark = gtk_file_system_memory_insert_bookmark;
   iface->remove_bookmark = gtk_file_system_memory_remove_bookmark;
   iface->list_bookmarks = gtk_file_system_memory_list_bookmarks;
+#if 0
+  iface->get_bookmark_label = gtk_file_system_memory_get_bookmark_label;
+  iface->set_bookmark_label = gtk_file_system_memory_set_bookmark_label;
+#endif
 }
 
 
@@ -496,8 +702,8 @@ gtk_file_system_memory_init(GtkFileSystemMemory *mfs)
   gtk_tree_store_set_column_types( GTK_TREE_STORE(mfs), G_N_ELEMENTS(types), types );
 
   mfs->bookmarks = NULL;
-  mfs->file_paths_to_be_deleted = NULL;
 
+  /* FIXME: override virtual functions instead */
   g_signal_connect( G_OBJECT(mfs), "row-inserted", 
                     G_CALLBACK(signal_row_inserted), NULL );
   g_signal_connect( G_OBJECT(mfs), "row-deleted",
@@ -616,7 +822,7 @@ gtk_file_system_memory_init(GtkFileSystemMemory *mfs)
 }
   #endif
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
 }
 
@@ -627,101 +833,185 @@ gtk_file_system_memory_finalize(GObject *object)
   GtkFileSystemMemory *mfs = GTK_FILE_SYSTEM_MEMORY( object );
   GObjectClass *object_class = G_OBJECT_CLASS(system_parent_class); 
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
+  g_slist_free(mfs->folders);
   g_slist_free(mfs->bookmarks);
   object_class->finalize( object );
 }
 
+static GSList *
+gtk_file_system_memory_path_find(GtkFileSystemMemory *fsm,
+                                 GtkTreePath *path)
+{
+  GSList *link;
+
+  for (link = fsm->folders; link != NULL; link = link->next)
+  {
+    GtkFileFolderMemory *ffm = link->data;
+    GtkTreePath *child_path;
+
+    /* The reference should always valid here. The invalidated references are
+       removed in "row-deleted" handler. */
+    child_path = gtk_tree_row_reference_get_path(ffm->data);
+    g_assert(child_path != NULL);
+
+    if (gtk_tree_path_compare(path, child_path) == 0)
+    {
+      gtk_tree_path_free(child_path);
+      return link;
+    }
+    gtk_tree_path_free(child_path);
+  }
+  return NULL;
+}
+
+static GtkFileFolderMemory *
+gtk_file_system_memory_path_lookup_parent(GtkFileSystemMemory *fsm,
+                                          GtkTreePath *path)
+{
+  GtkTreePath *folder_path = gtk_tree_path_copy(path);
+  GSList *link;
+
+  /* Move to parent folder's path */
+  if (!gtk_tree_path_up(folder_path))
+  {
+    gtk_tree_path_free(folder_path);
+    g_return_val_if_reached(NULL);
+  }
+
+  /* Lookup the parent folder. It exists if get_folder() has been called
+     for it. */
+  link = gtk_file_system_memory_path_find(fsm, folder_path);
+  gtk_tree_path_free(folder_path);
+
+  return link != NULL ? link->data : NULL;
+}
 
 void signal_row_inserted( GtkTreeModel *treemodel, GtkTreePath *path,
                           GtkTreeIter *iter, gpointer data)
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  GtkFileSystemMemory *fsm = GTK_FILE_SYSTEM_MEMORY(treemodel);
+  GtkFileFolderMemory *ffm = NULL;
+  GtkFilePath *file_path = NULL;
+  GSList *paths = NULL;
+  gchar *name;
+
+  DTEXT( __FUNCTION__ );
+
+  if( gtk_tree_path_get_depth( path ) == 1 )
+  {
+    g_signal_emit_by_name( GTK_FILE_SYSTEM(treemodel), "volumes-changed" );
+    return;
+  }
+
+  ffm = gtk_file_system_memory_path_lookup_parent(fsm, path);
+  if (ffm == NULL)
+    return;
+
+  gtk_tree_model_get(treemodel, iter, GTK_FILE_SYSTEM_MEMORY_COLUMN_NAME,
+                     &name, -1 );
+  if (name == NULL)
+  {
+    /* Node inserted without initial values. Don't show it until its values
+       are set (we get "row-changed" signal). This way is deprecated. */
+    ffm->appended = TRUE;
+  }
+  else
+  {
+    file_path =
+      gtk_file_system_memory_tree_path_to_file_path(GTK_FILE_SYSTEM(treemodel),
+                                                    path);
+    paths = g_slist_append(NULL, (gchar *)file_path);
+    g_signal_emit_by_name(ffm, "files-added", paths);
+    g_slist_free(paths);
+    g_free(file_path);
+    g_free(name);
+  }
 }
 
 void signal_row_changed( GtkTreeModel *treemodel, GtkTreePath *path,
                           GtkTreeIter *iter, gpointer data)
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  GtkFileSystemMemory *fsm = GTK_FILE_SYSTEM_MEMORY(treemodel);
+  GtkFileFolderMemory *ffm = NULL;
+  GtkFilePath *file_path = NULL;
+  GSList *paths = NULL;
+
+  DTEXT( __FUNCTION__ );
 
   if( gtk_tree_path_get_depth( path ) == 1 )
+  {
     g_signal_emit_by_name( GTK_FILE_SYSTEM(treemodel), "volumes-changed" );
+    return;
+  }
+
+  ffm = gtk_file_system_memory_path_lookup_parent(fsm, path);
+  if (ffm == NULL)
+    return;
+
+  /* Note that we can't handle filename changes here. Only content changes.
+     Filename changes would have to send "files-removed" signal, but we don't
+     have the old name for that anymore. API caller is expected to do renames
+     with remove+add. */
+  file_path =
+    gtk_file_system_memory_tree_path_to_file_path(GTK_FILE_SYSTEM(treemodel),
+                                                  path);
+  paths = g_slist_append(NULL, (gchar *)file_path);
+  if (!ffm->appended)
+    g_signal_emit_by_name(ffm, "files-changed", paths);
+  else
+  {
+    /* Initial values for added node were set only now. This way is
+       deprecated. */
+    ffm->appended = FALSE;
+    g_signal_emit_by_name(ffm, "files-added", paths);
+  }
+  g_slist_free(paths);
+  g_free(file_path);
 }
 
 void signal_row_deleted( GtkTreeModel *treemodel, GtkTreePath *path,
                          gpointer data)
 {
-  DTEXT( ":"__FUNCTION__"\n" );
-}
+  GtkFileSystemMemory *fsm = GTK_FILE_SYSTEM_MEMORY(treemodel);
+  GtkFileFolderMemory *ffm = NULL;
+  GSList *link = NULL, *next = NULL;
 
-
-void
-signal_row_reference_inserted( GtkTreeModel *treemodel, GtkTreePath *path,
-                               GtkTreeIter *iter, GtkFileFolderMemory *ffm )
-{
-  DTEXT( ":"__FUNCTION__"\n" );
-  ffm->appended = TRUE;
-}
-
-void
-signal_row_reference_changed( GtkTreeModel *treemodel, GtkTreePath *path,
-                              GtkTreeIter *iter, GtkFileFolderMemory *ffm )
-{
-  GtkTreePath *folder_path = NULL;
-  GtkTreeRowReference *row = ffm->data;
-
-  DTEXT( ":"__FUNCTION__"\n" );
-
-  if( !gtk_tree_row_reference_valid(row) )
+  if (fsm->deleted_file_path == NULL)
+  {
+    g_error("Use gtk_file_system_remove() instead of gtk_tree_store_remove()");
     return;
+  }
 
-  folder_path = gtk_tree_row_reference_get_path( row );
-  
-  if( gtk_tree_path_is_ancestor( folder_path, path ) && 
-    (gtk_tree_path_get_depth(path) == gtk_tree_path_get_depth(folder_path) + 1))
+  /* remove all invalid row references. there's as many of them as there are
+     deleted folders. */
+  for (link = fsm->folders; link != NULL; link = next)
+  {
+    GtkFileFolderMemory *link_ffm = link->data;
+
+    next = link->next;
+    if (!gtk_tree_row_reference_valid(link_ffm->data))
+    {
+      fsm->folders = g_slist_delete_link(fsm->folders, link);
+    }
+  }
+
+  if( gtk_tree_path_get_depth( path ) == 1 )
+  {
+    g_signal_emit_by_name( GTK_FILE_SYSTEM(treemodel), "volumes-changed" );
+    return;
+  }
+
+  /* notify parent folder that it lost files */
+  ffm = gtk_file_system_memory_path_lookup_parent(fsm, path);
+  if (ffm != NULL)
   {
     GSList *paths = NULL;
-    paths = g_slist_append( paths, 
-                          (gchar*)gtk_file_system_memory_tree_path_to_file_path( 
-                          GTK_FILE_SYSTEM(treemodel), path ) );
-    if( ffm->appended )
-    {
-      g_signal_emit_by_name(ffm, "files-added", paths);
-      ffm->appended = FALSE;
-    }
-    else
-      g_signal_emit_by_name( ffm, "files-changed", paths );
 
-    g_slist_free( paths );
-  }
-}
-
-void
-signal_row_reference_deleted( GtkTreeModel *treemodel, GtkTreePath *path,
-                              GtkFileFolderMemory *ffm )
-{
-  GtkTreePath *folder_path = NULL;
-  GtkTreePath *row_path = NULL;
-  GtkTreeRowReference *row = ffm->data;
-  GtkFileSystemMemory *fsm = GTK_FILE_SYSTEM_MEMORY(treemodel);
-
-  DTEXT( ":"__FUNCTION__"\n" );
-
-  if( !fsm->parent_path || !fsm->file_paths_to_be_deleted )
-    return;
-
-  folder_path = fsm->parent_path;
-  row_path = gtk_tree_row_reference_get_path( row );
-
-  if( !gtk_tree_path_compare(folder_path, row_path) )
-  {
-    GSList *paths = fsm->file_paths_to_be_deleted;
-    g_signal_emit_by_name( ffm, "files-removed", paths );
-    g_slist_free( paths );
-    fsm->file_paths_to_be_deleted = NULL;
-    gtk_tree_path_free( folder_path );
-    fsm->parent_path = NULL;
+    paths = g_slist_append(NULL, (gchar *)fsm->deleted_file_path);
+    g_signal_emit_by_name(ffm, "files-removed", paths);
+    g_slist_free(paths);
   }
 }
 
@@ -754,18 +1044,40 @@ gtk_file_folder_memory_iface_init( GtkFileFolderIface *iface )
 static void
 gtk_file_folder_memory_finalize( GObject *object )
 {
-  GtkTreePath *path = NULL;
   GtkFileFolderMemory *folder = GTK_FILE_FOLDER_MEMORY( object );
+  GtkFileSystemMemory *fsm = GTK_FILE_SYSTEM_MEMORY(folder->model);
+  GtkTreePath *path;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
-  path = gtk_tree_row_reference_get_path( folder->data );
-  gtk_tree_path_free( path );
+  /* If reference is invalid, it means we already removed it from folders tree
+     in gtk_file_system_memory_remove() */
+  path = gtk_tree_row_reference_get_path(folder->data);
+  if (path != NULL)
+  {
+    GSList *link;
+
+    link = gtk_file_system_memory_path_find(fsm, path);
+    g_assert(link != NULL);
+    fsm->folders = g_slist_delete_link(fsm->folders, link);
+
+    gtk_tree_path_free(path);
+  }
   gtk_tree_row_reference_free( folder->data );
 }
 
-
-
+#if 0
+static GtkFileSystemHandle *
+gtk_file_system_memory_get_info (GtkFileSystem                *file_system,
+				 const GtkFilePath            *path,
+				 GtkFileInfoType               types,
+				 GtkFileSystemGetInfoCallback  callback,
+				 gpointer                      data)
+{
+  DTEXT ( __FUNCTION__ );
+  return NULL;
+}
+#endif
 
 static GtkFileInfo *
 gtk_file_folder_memory_get_info( GtkFileFolder *folder, const GtkFilePath *path,
@@ -778,11 +1090,12 @@ gtk_file_folder_memory_get_info( GtkFileFolder *folder, const GtkFilePath *path,
   gint64 mod_time = 0;
   gint64 size = 0;
   gchar *mime = NULL;
+  GdkPixbuf *pbuf = NULL;
   GtkTreePath *tree_path = NULL;
   GtkFileInfo *info = NULL;
   GtkFileFolderMemory *ffm = GTK_FILE_FOLDER_MEMORY(folder);
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( path == NULL )
   {
@@ -815,7 +1128,9 @@ gtk_file_folder_memory_get_info( GtkFileFolder *folder, const GtkFilePath *path,
                       GTK_FILE_SYSTEM_MEMORY_COLUMN_NAME, &name,
                       GTK_FILE_SYSTEM_MEMORY_COLUMN_MOD_TIME, &mod_time,
                       GTK_FILE_SYSTEM_MEMORY_COLUMN_MIME, &mime,
-                      GTK_FILE_SYSTEM_MEMORY_COLUMN_SIZE, &size, -1 );
+                      GTK_FILE_SYSTEM_MEMORY_COLUMN_SIZE, &size,
+		      GTK_FILE_SYSTEM_MEMORY_COLUMN_ICON, &pbuf,
+		      -1 );
 
   info = gtk_file_info_new();
   gtk_file_info_set_display_name( info, name );
@@ -824,10 +1139,12 @@ gtk_file_folder_memory_get_info( GtkFileFolder *folder, const GtkFilePath *path,
   gtk_file_info_set_mime_type( info, "x-directory/normal" );
   gtk_file_info_set_modification_time( info, mod_time );
   gtk_file_info_set_size( info, size );
+  gtk_file_info_set_icon_pixbuf (info, pbuf);
   gtk_file_info_get_display_key( info );
   
   g_free( name );
   g_free( mime );
+  gdk_pixbuf_unref (pbuf);
   gtk_tree_path_free( tree_path );
 
   return info;
@@ -845,7 +1162,7 @@ gtk_file_folder_memory_list_children( GtkFileFolder *folder, GSList **children,
   GtkFileFolderMemory *ffm = GTK_FILE_FOLDER_MEMORY(folder);
   GtkTreeRowReference *row = ffm->data;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   *children = NULL;
 
@@ -895,7 +1212,7 @@ gtk_file_folder_memory_list_children( GtkFileFolder *folder, GSList **children,
 static gboolean
 gtk_file_folder_memory_is_finished_loading( GtkFileFolder *folder )
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
   return TRUE;
 }
 
@@ -911,7 +1228,7 @@ gtk_file_system_memory_list_volumes( GtkFileSystem *file_system )
   GtkTreeRowReference *row = NULL;
   GtkTreeModel *model = GTK_TREE_MODEL( file_system );
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( !gtk_tree_model_get_iter_first( model, &iter ) )
     return NULL;
@@ -939,7 +1256,7 @@ gtk_file_system_memory_get_volume_for_path( GtkFileSystem *file_system,
   gchar **first = NULL;
   gchar *str = NULL;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   g_return_val_if_fail( path, NULL );
   
@@ -976,65 +1293,75 @@ gtk_file_system_memory_get_volume_for_path( GtkFileSystem *file_system,
   return (GtkFileSystemVolume*)row;
 }
 
-/* FIXME: Some kind of cache (=hashmap) for created folders would be nice */
-static GtkFileFolder *
-gtk_file_system_memory_get_folder( GtkFileSystem *file_system,
-				                           const GtkFilePath *path,
-                                   GtkFileInfoType types, GError **error )
+static GtkFileSystemHandle *
+gtk_file_system_memory_get_folder (GtkFileSystem *file_system,
+				   const GtkFilePath *path,
+                                   GtkFileInfoType types,
+				   GtkFileSystemGetFolderCallback callback,
+				   gpointer data)
 {
+  GtkFileSystemMemory *fsm = GTK_FILE_SYSTEM_MEMORY(file_system);
   GtkFileFolderMemory *ffm;
   GtkTreeModel *model;
   GtkTreePath *tree_path;
+  GError *error = NULL;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
-  if( !(tree_path = gtk_file_system_memory_file_path_to_tree_path( file_system,
-                                                                   path )) )
-  {
-    g_set_error( error, GTK_FILE_SYSTEM_ERROR, 
-                 GTK_FILE_SYSTEM_ERROR_NOT_FOLDER,
-                 "Path %s is not valid", (gchar*)path );
-    return NULL;
-  }
+  tree_path = gtk_file_system_memory_file_path_to_tree_path (file_system,
+							     path);
+  if(tree_path == NULL)
+    {
+      g_set_error (&error, GTK_FILE_SYSTEM_ERROR, 
+		   GTK_FILE_SYSTEM_ERROR_NOT_FOLDER,
+		   "Path %s is not valid", (gchar*)path );
+      return queue_get_folder_callback (file_system, callback,
+					NULL, error, data);
+    }
 
   model = GTK_TREE_MODEL(file_system);
-  
+
+  /* FIXME: Reuse earlier created folder object if it exists. */  
   ffm = g_object_new( GTK_TYPE_FILE_FOLDER_MEMORY, NULL );
   ffm->data = gtk_tree_row_reference_new( model, tree_path );
   ffm->model = model;
 
-  g_signal_connect_object( G_OBJECT(model), "row-changed",
-                    G_CALLBACK(signal_row_reference_changed), ffm, 0 );
-  g_signal_connect_object( G_OBJECT(model), "row-inserted",
-                    G_CALLBACK(signal_row_reference_inserted), ffm, 0 );
-  g_signal_connect_object( G_OBJECT(model), "row-deleted",
-                    G_CALLBACK(signal_row_reference_deleted), ffm, 0 );
+  /* Pointer is removed from this list in folder finalize */
+  fsm->folders = g_slist_prepend(fsm->folders, ffm);
 
-  gtk_tree_path_free( tree_path );
-                    
-  return GTK_FILE_FOLDER(ffm);
+  gtk_tree_path_free (tree_path);
+  
+  return queue_get_folder_callback (file_system, callback,
+				    GTK_FILE_FOLDER (ffm),
+				    NULL, data);
 }
 
 
-static gboolean
-gtk_file_system_memory_create_folder( GtkFileSystem *file_system,
-					                            const GtkFilePath *path, GError **error)
+static GtkFileSystemHandle *
+gtk_file_system_memory_create_folder (GtkFileSystem *file_system,
+				      const GtkFilePath *path,
+				      GtkFileSystemCreateFolderCallback callback,
+				      gpointer data)
 {
   gint path_len = 0, slash = 0, len = 0;
   gchar *str = NULL, *current = NULL, *to_path = NULL, *new_folder = NULL;
   GtkTreePath *tree_path = NULL;
   GtkTreeStore *store;
   GtkTreeIter iter, ipath;
+  GError *error = NULL;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
+
+  /* FIXME: This function is fuzzy, rework needed... */
 
   if( (tree_path = gtk_file_system_memory_file_path_to_tree_path( file_system, 
                                                                   path )) )
   {
-    g_set_error( error, GTK_FILE_SYSTEM_ERROR, 
+    g_set_error (&error, GTK_FILE_SYSTEM_ERROR, 
                  GTK_FILE_SYSTEM_ERROR_ALREADY_EXISTS,
-                 "Folder %s already exists", (gchar*)path );
-    return FALSE;
+                 "Folder %s already exists", (gchar*)path);
+    return queue_create_folder_callback (file_system, callback,
+					 path, error, data);
   }
   
   str = (gchar*)path;
@@ -1064,12 +1391,13 @@ gtk_file_system_memory_create_folder( GtkFileSystem *file_system,
   if( !(tree_path = gtk_file_system_memory_file_path_to_tree_path(file_system,
        (GtkFilePath*)to_path)) )
   {
-    g_set_error( error, GTK_FILE_SYSTEM_ERROR, 
+    g_set_error (&error, GTK_FILE_SYSTEM_ERROR, 
                  GTK_FILE_SYSTEM_ERROR_BAD_FILENAME,
                  "Path to %s is not valid", (gchar*)path );
     g_free( new_folder );
     g_free( to_path );             
-    return FALSE;
+    return queue_create_folder_callback (file_system, callback,
+					 path, error, data);
   }
 
   store = GTK_TREE_STORE( file_system );
@@ -1089,7 +1417,9 @@ gtk_file_system_memory_create_folder( GtkFileSystem *file_system,
   g_free( to_path );
   g_free( new_folder );
   gtk_tree_path_free( tree_path );
-  return TRUE;
+
+  return queue_create_folder_callback (file_system, callback,
+				       path, NULL, data);
 }
 
 
@@ -1097,7 +1427,7 @@ static void
 gtk_file_system_memory_volume_free( GtkFileSystem *file_system,
 				                            GtkFileSystemVolume *volume )
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
   gtk_tree_row_reference_free( (GtkTreeRowReference*)volume );
 }
 
@@ -1106,7 +1436,7 @@ static GtkFilePath *
 gtk_file_system_memory_volume_get_base_path( GtkFileSystem *file_system,
 						                                 GtkFileSystemVolume  *volume )
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
   return (GtkFilePath*)gtk_file_system_memory_volume_get_display_name(
                                                          file_system, volume );
 }
@@ -1116,20 +1446,22 @@ static gboolean
 gtk_file_system_memory_volume_get_is_mounted( GtkFileSystem  *file_system,
                                               GtkFileSystemVolume  *volume )
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
   return TRUE;
 }
 
 
-static gboolean
-gtk_file_system_memory_volume_mount( GtkFileSystem *file_system,
-					                           GtkFileSystemVolume *volume,
-                                     GError **error )
+static GtkFileSystemHandle *
+gtk_file_system_memory_volume_mount (GtkFileSystem *file_system,
+				     GtkFileSystemVolume *volume,
+				     GtkFileSystemVolumeMountCallback callback,
+				     gpointer data)
 {
-  DTEXT( ":"__FUNCTION__"\n" );
-  /* DONTKNOW - Should we emit a signal or not? */
-  /*g_signal_emit_by_name (file_system, "volumes-changed", NULL);*/
-  return TRUE;
+  DTEXT( __FUNCTION__ );
+  /* Volumes are always mounted in memory filesystem, so there's no need to
+     do anything here. */
+  return queue_volume_mount_callback (file_system, callback,
+				      volume, NULL, data);
 }
 
 
@@ -1143,7 +1475,7 @@ gtk_file_system_memory_volume_get_display_name( GtkFileSystem *file_system,
   GtkTreeModel *model;
   GtkTreeRowReference *row = (GtkTreeRowReference*)volume;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( !(path = gtk_tree_row_reference_get_path(row)) )
     return NULL;
@@ -1163,7 +1495,7 @@ gtk_file_system_memory_volume_get_display_name( GtkFileSystem *file_system,
   return str;
 }
 
-
+#if 0
 static GdkPixbuf *
 gtk_file_system_memory_volume_render_icon( GtkFileSystem *file_system,
 					                                 GtkFileSystemVolume *volume,
@@ -1175,7 +1507,7 @@ gtk_file_system_memory_volume_render_icon( GtkFileSystem *file_system,
   GtkTreePath *tree_path = gtk_tree_row_reference_get_path(
       					   (GtkTreeRowReference*)volume );
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( !tree_path )
   {
@@ -1194,7 +1526,7 @@ gtk_file_system_memory_volume_render_icon( GtkFileSystem *file_system,
   gtk_tree_path_free( tree_path );
   return pbuf;
 }
-
+#endif
 
 static gboolean
 gtk_file_system_memory_get_parent( GtkFileSystem *file_system,
@@ -1203,7 +1535,7 @@ gtk_file_system_memory_get_parent( GtkFileSystem *file_system,
 {
   GtkTreePath *tree_path = NULL;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( !(tree_path = gtk_file_system_memory_file_path_to_tree_path( file_system,
                                                                    path )) )
@@ -1241,7 +1573,7 @@ gtk_file_system_memory_make_path( GtkFileSystem *file_system,
   GtkFilePath *result = NULL;
   const gchar *path = gtk_file_path_get_string(base_path);
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   len = strlen(path)-1;
 
@@ -1266,7 +1598,7 @@ gtk_file_system_memory_parse( GtkFileSystem *file_system,
 {
   GtkTreePath *tree_path = NULL;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( !(tree_path = gtk_file_system_memory_file_path_to_tree_path( file_system,
                                                                    base_path)) )
@@ -1289,7 +1621,7 @@ static gchar *
 gtk_file_system_memory_path_to_uri( GtkFileSystem *file_system,
 				                            const GtkFilePath *path )
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
   return g_filename_to_uri( gtk_file_path_get_string(path), NULL, NULL );
 }
 
@@ -1298,7 +1630,7 @@ static gchar *
 gtk_file_system_memory_path_to_filename( GtkFileSystem *file_system,
 					                               const GtkFilePath *path )
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
   return g_strdup( gtk_file_path_get_string(path) );
 }
 
@@ -1310,7 +1642,7 @@ gtk_file_system_memory_uri_to_path( GtkFileSystem *file_system,
   GtkFilePath *path = NULL;
   gchar *filename = strchr( uri, ':' );
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if (filename)
     {
@@ -1329,11 +1661,11 @@ static GtkFilePath *
 gtk_file_system_memory_filename_to_path( GtkFileSystem *file_system,
 					                               const gchar *filename )
 {
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
   return gtk_file_path_new_dup(filename);
 }
 
-
+#if 0
 static GdkPixbuf *
 gtk_file_system_memory_render_icon( GtkFileSystem *file_system,
 				                            const GtkFilePath *path, GtkWidget *widget,
@@ -1344,7 +1676,7 @@ gtk_file_system_memory_render_icon( GtkFileSystem *file_system,
   GtkTreeModel *model;
   GtkTreePath *tree_path;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( !(tree_path = gtk_file_system_memory_file_path_to_tree_path(
        file_system, path )) )
@@ -1366,7 +1698,7 @@ gtk_file_system_memory_render_icon( GtkFileSystem *file_system,
   
   return pbuf;
 }
-
+#endif
 
 static gboolean
 gtk_file_system_memory_insert_bookmark( GtkFileSystem *file_system,
@@ -1376,7 +1708,7 @@ gtk_file_system_memory_insert_bookmark( GtkFileSystem *file_system,
   GtkTreePath *tree_path = NULL;
   GtkFileSystemMemory *fsm = GTK_FILE_SYSTEM_MEMORY(file_system);
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   tree_path = gtk_file_system_memory_file_path_to_tree_path( file_system, path );
   
@@ -1405,7 +1737,7 @@ gtk_file_system_memory_remove_bookmark( GtkFileSystem *file_system,
   GSList *current = NULL;
   start = current = GTK_FILE_SYSTEM_MEMORY(file_system)->bookmarks;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   while( current )
   {
@@ -1433,7 +1765,7 @@ gtk_file_system_memory_list_bookmarks( GtkFileSystem *file_system )
   GSList *current = GTK_FILE_SYSTEM_MEMORY(file_system)->bookmarks;
   GSList *new = NULL;
 
-  DTEXT( ":"__FUNCTION__"\n" );
+  DTEXT( __FUNCTION__ );
 
   if( current )
     do
